import numpy
from mpi4py import MPI

from matrix import Matrix


class MatrixMPI:
    __mpi_comm__ = None
    __mpi_size__ = None
    __mpi_rank__ = None

    __matrix__: Matrix = None
    __chunk__: list = None

    def __init__(self, mpi_comm, data=None, shape=None, structure=None, model=None, offset=None, n=None):
        self.__mpi_comm__ = mpi_comm
        self.__mpi_size__ = self.__mpi_comm__.Get_size()
        self.__mpi_rank__ = self.__mpi_comm__.Get_rank()

        self.__matrix__ = Matrix(data=data, shape=shape, structure=structure, model=model, offset=offset, n=n)

        total_amount_of_rows = self.__matrix__.shape()[0]
        chunks = numpy.array_split(list(range(total_amount_of_rows)), self.__mpi_size__)
        self.__chunk__ = chunks[self.__mpi_rank__].tolist()

    @staticmethod
    def of(matrix: Matrix):
        return MatrixMPI(matrix.get_data(), matrix.shape())

    def __str__(self):
        return str(self.__matrix__)

    def get_rank_submatrix(self):
        rows = len(self.__chunk__)
        cols = self.__matrix__.shape()[1]
        return Matrix(self.__matrix__[self.__chunk__], (rows, cols))

    def transpose(self):
        """
        :return: the transpose of the matrix
        """
        return MatrixMPI.of(self.__matrix__.transpose())

    def T(self):
        """
        Same as ``matrix.transpose()``

        :return: see ``matrix.transpose()``
        """
        return self.transpose()

    def __eq__(self, other):
        """
        Return ``self==value``

        :param other: The object to compare to; must be either a ``MatrixMPI``, ``Matrix``, a ``list`` or a ``numpy.ndarray``
        :return: True if data in the matrix are equal to the given data in other for each component, otherwise False
        """
        if isinstance(other, MatrixMPI):
            return self.__matrix__ == other.__matrix__
        else:
            return self.__matrix__ == other

    def __neg__(self):
        gathered_data = self.__mpi_comm__.gather(-self.get_rank_submatrix())
        data = self.__mpi_comm__.bcast(gathered_data)
        return MatrixMPI.of(Matrix.flatten(data))

    def __add__(self, other):
        if isinstance(other, MatrixMPI):
            other = other.get_rank_submatrix()
        gathered_data = self.__mpi_comm__.gather(self.get_rank_submatrix() + other)
        data = self.__mpi_comm__.bcast(gathered_data)
        return MatrixMPI.of(Matrix.flatten(data))

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return -self + other

    def __truediv__(self, other):
        gathered_data = self.__mpi_comm__.gather(self.get_rank_submatrix() / other)
        data = self.__mpi_comm__.bcast(gathered_data)
        return MatrixMPI.of(Matrix.flatten(data))

    def __mul__(self, other):
        if isinstance(other, MatrixMPI):
            return self.__matrix__ * other.__matrix__
        else:
            return self.__matrix__ * other

    def __rmul__(self, other):
        return self * other

    def norm(self, f: str = "frobenius"):
        """
        Calculates the norm of the matrix.

        A norm is a positive definit, absolute homogeneous and subadditive function.
        For Matrices a norm is also sub-multiplicative.

        :param f: The norm to be used, could be either "frobenius", "row sum" or "col sum"

        :return: the norm as a number
        """
        return self.__matrix__.norm(f)

    def __getitem__(self, key):
        return self.__matrix__[key]

    def __setitem__(self, key, value):
        self.__matrix__[key] = value

